/* ************************************************************************ */
/*                                                                          */
/*                UTILITIES FOR FAST FOURIER TRANSFORM METHODS              */   
/*                                                                          */
/* ************************************************************************ */

// IMPORT JTRANSFORMS (IN code DIRECTORY OF SKETCH
import edu.emory.mathcs.jtransforms.fft.*;
import edu.emory.mathcs.utils.*;
import edu.emory.mathcs.jtransforms.dst.*;
import edu.emory.mathcs.jtransforms.dht.*;
import edu.emory.mathcs.jtransforms.dct.*;

// Convolution
float[][] runConvolution(float[][] imageSpectrum, float[][] kernelSpectrum){
  // Find the complexproduct of the spectra
  float[][] tmp = complexMultiply(imageSpectrum,kernelSpectrum);
  // The inverse Fourier transform yields the filtered image
  fft.complexInverse(tmp,true);
  return getRealPart(tmp);
}

// Given a float data buffer create a spectrum
float[][] imageSpectrum(float[][] data){
  // Compute the Fourier spectrum of the image. This yields a complex spectrum. Re and Im parts are interlaced!
  float[][] spectrum = new float[SZ][2*SZ];
  for (int i=0; i<SZ; i++) for(int j=0; j<SZ; j++) {
    spectrum[i][j]    = data[i][j]; // Re
    spectrum[i][j+SZ] =        0.0; // Im
  }
  fft.realForwardFull(spectrum); // spectrum contains the image spectrum
  return spectrum;
}

float[][] kernelSpectrum(float sigma){
 float[][] kernel = new float[SZ][SZ];
 for (int i = 0; i < SZ; i++) for (int j = 0; j < SZ; j++) {
  float u = (float) i;
  float v = (float) j;
  if (i > SZ/2) u = u - SZ;
  if (j > SZ/2) v = v - SZ;
  kernel[i][j] = gaussian2D(u,v,sigma);
 }
 float[][] spectrum = new float[SZ][2*SZ];
 for (int i=0; i<SZ; i++) for(int j=0; j<SZ; j++) { 
   spectrum[i][j]     = kernel[i][j]; // Re
   spectrum[i][j+SZ] = 0.0;           // Im
 }
 // The transform is complex, Re and Im interlaced
 fft.realForwardFull(spectrum);
 return spectrum;
}

float[][] kernelSpectrum(int xOrder, int yOrder, float sigma){
  // Construct the kernel
  // This is a real function
  float[][] kernel = new float[SZ][SZ];
  float[]   column = new float[SZ];
  float[]      row = new float[SZ];
  for (int i=0; i<SZ; i++) {
    float u = (float) i;
    if (i > SZ/2) u = u - SZ;
    column[i] = pow( -1.0/(sigma*sqrt(2.0)),xOrder) * hermitePolynomial(xOrder,u/(sigma*sqrt(2))) * gaussian(u,sigma);
  }
  for(int j=0; j<SZ; j++) {
    float v = (float) j;
    if (j > SZ/2) v = v - SZ;
    row[j] = pow( -1.0/(sigma*sqrt(2.0)),yOrder) * hermitePolynomial(yOrder,v/(sigma*sqrt(2))) * gaussian(v,sigma);
  }
  for (int i = 0; i < SZ; i++) for (int j = 0; j < SZ; j++) kernel[i][j] = column[i] * row[j];
  // Fourier transform it, creating a new array if none exists, otherwise overwrite
  // The transform is complex, Re and Im interlaced
  // The kernel itself is simply filled up with zeroes
  float[][] spectrum = new float[SZ][2*SZ];
  for (int i=0; i<SZ; i++) for(int j=0; j<SZ; j++) { 
      spectrum[i][j]    = kernel[i][j]; // Re
      spectrum[i][j+SZ] = 0.0;          // Im
    }
  fft.realForwardFull(spectrum);
  return spectrum;
}

// Returns the product of two complex spectra a & b
// Re and Im are interlaced!
float[][] complexMultiply(float[][] a, float[][] b) {
  float[][] result = new float[SZ][2*SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    result[i][2*j]   = a[i][2*j] * b[i][  2*j] - a[i][2*j+1] * b[i][2*j+1];
    result[i][2*j+1] = a[i][2*j] * b[i][2*j+1] + a[i][2*j+1] * b[i][  2*j];
  }
  return result;
}

// This picks out the real part and ignores the imaginary part
// Re and Im are interlaced!
float[][] getRealPart(float[][] aComplexSpectrum) {
  float[][] result = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) result[i][j] = aComplexSpectrum[i][2*j];
  return result;
}

// 1D Gaussian
float gaussian(float x, float sigma) {return exp(-sq(x)/(2f*sq(sigma)))/(sqrt(2f*PI)*sigma);}

// 2D Gaussian
float gaussian2D(float x, float y, float sigma) {return exp(-(sq(x)+sq(y))/(2f*sq(sigma)))/(2f*PI*sq(sigma));}

// Hermite polynomials, via the recursion relation
float hermitePolynomial(int n, float x) {
  float result;
  switch (n) {
    case(0):  result = 1.0;
              break;
    case(1):  result = 2.0 * x;
              break;
    case(2):  result = 4.0 * sq(x) - 2.0;
              break;
    case(3):  result = 8.0 * pow(x,3) - 12.0 * x;
              break;
    case(4):  result = 16.0 * pow(x,4) - 48.0 * sq(x) + 12.0;
              break;
    default:  result = 2.0*(x*hermitePolynomial(n-1,x)-(n-1)*hermitePolynomial(n-2,x));
  }
  return result;
}