/* ************************************************************************ */
/*                                                                          */
/*                         GENERATION OF RANDOM NUMBERS                     */   
/*                         USES A QUALITY PCG ALGORITHM                     */
/*                                                                          */
/* ************************************************************************ */

import java.util.concurrent.atomic.AtomicLong;

class Pcg32 {
  long state;
  long inc;

  final static long MULTIPLIER = 6364136223846793005L;

  Pcg32(long initState, long initSeq) {
      seed(initState, initSeq);
  }

  Pcg32() { seed();}

  // Returns the next pseudorandom, approximately uniformly distributed int
  // value from this random number generator's sequence. 
  int nextInt() {
      long oldState = state;
      state = oldState * MULTIPLIER + inc;
      int xorShifted = (int) (((oldState >>> 18) ^ oldState) >>> 27);
      int rot = (int) (oldState >>> 59);
      return Integer.rotateRight(xorShifted, rot);
  }
   
  // Returns the next pseudorandom, approximately uniformly distributed int
  // value between 0 (inclusive) and {@code n} (exclusive).
  int nextInt(int n) {
      if (n <= 0) {
          throw new IllegalArgumentException("n must be positive");
      }
      // the special treatment of powers of 2 as in Random.nextInt(int) shouldn't be necessary
      int bits, val;
      do {
          bits = nextInt() >>> 1;
          val = bits % n;
      } while (bits - val + (n-1) < 0);
      return val;
  }

   // Returns the next pseudorandom, approximately uniformly distributed boolean value.
  boolean nextBoolean() { return (nextInt() & 1) != 0;}

   // Returns the next pseudorandom, approximately uniformly distributed float value between 0.0 (inclusive)
   // and 1.0 (exclusive)
 float nextFloat() { return nextBits(24) / ((float)(1 << 24));}

   // Returns the next pseudorandom, approximately uniformly distributed float value between 0.0 (inclusive)
   // and bound (exclusive).
  float nextFloat(float bound) { return bound * nextFloat(); }
  
      int nextBits(int bits) {return nextInt() >>> (32 - bits);}
  
  float nextNextGaussian;
  boolean haveNextNextGaussian = false;

   // Returns the next pseudorandom, Gaussian ("normally") distributed
   // float value with mean 0.0 and standard
   // deviation 1.0 from this random number generator's sequence.
  float nextGaussian() {
      // See Knuth, ACP, Section 3.4.1 Algorithm C.
      if (haveNextNextGaussian) {
          haveNextNextGaussian = false;
          return nextNextGaussian;
      } else {
          float v1, v2, s;
          do {
              v1 = 2 * nextFloat() - 1; // between -1 and 1
              v2 = 2 * nextFloat() - 1; // between -1 and 1
              s = v1 * v1 + v2 * v2;
          } while (s >= 1 || s == 0);
          float multiplier = sqrt(-2.0 * log(s)/s);
          nextNextGaussian = v2 * multiplier;
          haveNextNextGaussian = true;
          return v1 * multiplier;
      }
  }

   // Returns the next pseudorandom, Gaussian ("normally") distributed
   // double value with given mean and standard
   // deviation from this random number generator's sequence.
  float nextGaussian(float mean, float standardDeviation) { return nextGaussian() * standardDeviation + mean;}

  void seed(long initState, long initSeq) {
      state = 0;
      inc = 2 * initSeq + 1;
      nextInt();
      state += initState;
      nextInt();
  }

   // Initializes the generator with current time as the state and
   // a (very likely to be) unique stream number. This ensures that
   // even if multiple generators are initialized using this method
   // while System#nanoTime() returns the same time, they
   // will have different sequences.
   
  // JUST A SANITY CHECK
   
  void seed() { seed(System.nanoTime(), streamUniquifier());}

  long streamUniquifier() {
      for (;;) {
          long current = streamUniquifier.get();
          long next = current * 181783497276652981L;
          if (streamUniquifier.compareAndSet(current, next)) return next;
      }
  }

  final AtomicLong streamUniquifier = new AtomicLong(System.identityHashCode(Pcg32.class));
}